/*
based on code of Copyright (c) Martin Kinkelin
*/

#pragma once

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string>


/// <summary>
/// Some helper functions for reading values from the registry.
/// </summary>
namespace Npstates{
class Registry
{
public:

	static bool GetCString(HKEY key, const char* value, char* result, ureg32 *prsize)
	{
		if (key == NULL)
			return false;

		return (RegQueryValueEx(key, value, NULL, NULL, (LPBYTE)result, (LPDWORD)prsize) == ERROR_SUCCESS);
	}

	static bool GetString1024(HKEY key, const char* value, std::string& result)
	{
		if (key == NULL)
			return false;

		char buffer[1024];
		DWORD size = 1024;

		if (RegQueryValueEx(key, value, NULL, NULL, (LPBYTE)buffer, &size) == ERROR_SUCCESS)
		{
			result = std::string(buffer, size);
			return true;
		}

		return false;
	}

	static bool GetDword(HKEY key, const char* value, DWORD& result)
	{
		if (key == NULL)
			return false;

		DWORD temp;
		DWORD size = sizeof(DWORD);
		if (RegQueryValueEx(key, value, NULL, NULL, (LPBYTE)&temp, &size) == ERROR_SUCCESS)
		{
			result = temp;
			return true;
		}

		return false;
	}

	static bool GetBool(HKEY key, const char* value, bool& result)
	{
		DWORD temp;
		if( GetDword(key, value, temp) )
		{
			result = (temp != 0);
			return true;
		}

		return false;
	}

};

//namespace Npstates
}
