/*
based on code of Copyright (c) Martin Kinkelin
*/

/*
 * Main module, tightly related to the BaseService class.
 */

#include "AMD_15h_MSR_PSTATES.h"

namespace Npstates{

bool is_debug=
#ifdef DEBUG
	true;
#else
	false;
#endif

//namespace Npstates
}
using namespace Npstates;

#include <signal.h>
#include <iostream>
using namespace std;

#define VER 0.2

// prototypes
void WINAPI ServiceMain(__in DWORD dwArgc, __in LPTSTR* lpszArgv);
DWORD WINAPI ServiceCtrlHandlerEx(__in DWORD dwControl, __in DWORD dwEventType,
	__in LPVOID lpEventData, __in LPVOID lpContext);

// single service instance
static ServiceType* _service = NULL;


namespace{
/// <summary>Ctrl+C handler.</summary>
void SigintHandler(int signal)
{
	if (_service != NULL)
		_service->Stop();
}

void	usage(const char *msg=0){
const char	*const prog_name="AMD_15h_MSR_PSTATES";
	if(msg)cout << "error: " << msg;
	cout << prog_name;
	cout << " V" << VER << "\n";
	cout << "Usage: { prog_name } {parameter}\n";
	cout << "parameters:\n\t\n-i\tinstall service\n\t-u\tuninstall service\n\t-d dump current MSRS {MSR_Lxx,MSR_Hxx}to file 'dump.a15'\n\t-h this help message\n";
	cout << "user setup:\n\tHKEY_LOCAL_MACHINE\\SOFTWARE\\AMD_15h_MSR_PSTATES\n";
	cout << "\t\tcur_set_name:REG_SZ point to sub_dir with MSR_C001_006x set\n";
	cout << "\t\t{MSR_Lxx:DWORD,MSR_Hxx:DWORD} (MSR_C001_006x set)\n";
	cout << "\t\tis_check_CPU:DWORD (0), is_mask_IDD:DWORD (1), is_mask_K10_NbVID:DWORD (1)\n";
	cout << "\t\tis_avoid_wr_cur_pstate:DWORD (0), is_sw_lowest_pstate:DWORD (0)\n";
	exit(1);
}

bool	is_dump=false;
//namespace
}


/// <summary>Entry point for the program.</summary>
int main(int argc, char* argv[])
{
	//help
	if (argc > 1 && _stricmp(argv[1], "-h") == 0)usage();

	//dump
	if (argc > 1 && _stricmp(argv[1], "-d") == 0){
		is_dump=true;
		is_debug=true;
	}

	if(is_debug){

	// if is_debug is defined, the program is started normally (not as a service)
	// the "service" is stopped by pressing Ctrl+C (SIGINT)
	signal(SIGINT, SigintHandler);

	ServiceMain(0, NULL);

	}else{

	// try to install the service if the "-i" switch is used
	if (argc > 1 && _stricmp(argv[1], "-i") == 0)
	{
		// try to connect to the local service control manager
		const SC_HANDLE scm = OpenSCManager(NULL, NULL, SC_MANAGER_CREATE_SERVICE);
		if (scm == NULL)
		{
			cerr << "Could not connect to the service control manager." << endl;
			return 1;
		}

		// get the path of this executable in double quotes
		char buffer[MAX_PATH + 2];
		buffer[0] = '\"';
		GetModuleFileName(NULL, buffer + 1, MAX_PATH);
		strcat_s(buffer, "\"");

		// try to open the existing service
		SC_HANDLE service = OpenService(scm, NAME, SERVICE_STOP | DELETE);
		if (service != NULL)
		{
			// try to stop and delete the service
			SERVICE_STATUS serviceStatus;
			if ((!ControlService(service, SERVICE_CONTROL_STOP, &serviceStatus) && GetLastError() != ERROR_SERVICE_NOT_ACTIVE) ||
				!DeleteService(service))
			{
				cerr << "Could not uninstall the service (SERVICE_CONTROL_STOP)." << endl;
				CloseServiceHandle(service);
				CloseServiceHandle(scm);
				return 4;
			}

			CloseServiceHandle(service);
			service = NULL;
		}

		// try to install the service
		service = CreateService(scm, NAME, DISPLAY_NAME, SERVICE_CHANGE_CONFIG,
			SERVICE_WIN32_OWN_PROCESS, START_TYPE, SERVICE_ERROR_NORMAL,
			buffer, NULL, NULL, NULL, NULL, "");
		if (service == NULL)
		{
			cerr << "Could not install the service." << endl;
			CloseServiceHandle(scm);
			return 2;
		}

		// attempt to set the description
		SERVICE_DESCRIPTION description;
		description.lpDescription = DESCRIPTION;
		ChangeServiceConfig2(service, SERVICE_CONFIG_DESCRIPTION, &description);

#if defined(ENABLE_DELAYED_AUTOSTART) && START_TYPE == SERVICE_AUTO_START
		// attempt to enable the autostart delay supported by Vista+
		SERVICE_DELAYED_AUTO_START_INFO delayedAutoStartInfo;
		delayedAutoStartInfo.fDelayedAutostart = TRUE;
		ChangeServiceConfig2(service, SERVICE_CONFIG_DELAYED_AUTO_START_INFO, &delayedAutoStartInfo);
#endif

		CloseServiceHandle(service);
		CloseServiceHandle(scm);

		return 0;
	}

	// try to uninstall the service if the "-u" switch is used
	if (argc > 1 && _stricmp(argv[1], "-u") == 0)
	{
		// try to connect to the local service control manager
		SC_HANDLE scm = OpenSCManager(NULL, NULL, SC_MANAGER_CONNECT);
		if (scm == NULL)
		{
			cerr << "Could not connect to the service control manager." << endl;
			return 1;
		}

		// try to open the service
		SC_HANDLE service = OpenService(scm, NAME, SERVICE_STOP | DELETE);
		if (service == NULL)
		{
			cerr << "Could not uninstall the service (SERVICE_STOP)." << endl;
			CloseServiceHandle(scm);
			return 3;
		}

		// try to stop and delete the service
		SERVICE_STATUS serviceStatus;
		if ((!ControlService(service, SERVICE_CONTROL_STOP, &serviceStatus) && GetLastError() != ERROR_SERVICE_NOT_ACTIVE) ||
		    !DeleteService(service))
		{
			cerr << "Could not uninstall the service (SERVICE_CONTROL_STOP)." << endl;
			CloseServiceHandle(service);
			CloseServiceHandle(scm);
			return 4;
		}

		CloseServiceHandle(service);
		CloseServiceHandle(scm);

		return 0;
	}

	// try to start the service
	SERVICE_TABLE_ENTRY dispatchTable[] = { { NAME, ServiceMain }, { 0, NULL } };
	if (!StartServiceCtrlDispatcher(dispatchTable))
		return 10;

	} // DEBUG

	// return when the service has terminated
	return 0;
}


/// <summary>Invoked when the service is started.</summary>
void WINAPI ServiceMain(__in DWORD dwArgc, __in LPTSTR* lpszArgv)
{
	// try to register the service control handler
	SERVICE_STATUS_HANDLE statusHandle = NULL;

	if(!is_debug){

	if ((statusHandle = RegisterServiceCtrlHandlerEx(NAME, ServiceCtrlHandlerEx, NULL)) == NULL)
		return;
	}

	// initialize the status
	SERVICE_STATUS status;
	status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	status.dwCurrentState = SERVICE_STOPPED;
	status.dwControlsAccepted = 0;
	status.dwWin32ExitCode = NO_ERROR;
	status.dwServiceSpecificExitCode = 0;
	status.dwCheckPoint = 0;
	status.dwWaitHint = 0;

	// try to create the service instance
	try
	{
		_service = new ServiceType(statusHandle, status);
	}
	catch (exception)
	{
		if(is_debug){
		SetServiceStatus(statusHandle, &status);
		}
		return;
	}

	// set our status to Running as soon as possible, because until then
	// the service control manager is blocked
	_service->UpdateStatus(SERVICE_RUNNING);

	// try to initialize the service
	if (!_service->OnInitialize())
	{
		_service->UpdateStatus(SERVICE_STOPPED);
		delete _service;
		_service = NULL;
		return;
	}

	//requested dump MSRs?
	if(is_dump){
	
	FILE *fo=fopen("dump.a15","wb");

	if(fo){
	for(ureg32 idx=0; idx<_service->MSRS_low.size(); ++idx ){
	
	_service->ReadMSR_C001_006x( 
			0, 
			idx, 
			_service->MSRS_low[idx],
			_service->MSRS_high[idx]
		);

	std::fprintf(fo,"reg idx %02u: low %08X; high %08X\n", idx, _service->MSRS_low[idx], _service->MSRS_high[idx].second);
	}
	
	fclose(fo);
	}}else{

	// start the service
	_service->UpdateStatus(SERVICE_RUNNING, SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_POWEREVENT);
	_service->OnRun();

	// wait until the service is stopped
	_service->Wait(INFINITE);

	_service->UpdateStatus(SERVICE_STOP_PENDING);
	_service->OnStop();
	}

	_service->UpdateStatus(SERVICE_STOPPED);

	delete _service;
	_service = NULL;
}

/// <summary>
/// Invoked asynchronously by the service control manager to respond to commands and notifications.
/// </summary>
DWORD WINAPI ServiceCtrlHandlerEx(__in DWORD dwControl, __in DWORD dwEventType,
	__in LPVOID lpEventData, __in LPVOID lpContext)
{
	switch (dwControl)
	{
		case SERVICE_CONTROL_STOP:
		{
			_service->Stop();
			return NO_ERROR;
		}

		case SERVICE_CONTROL_POWEREVENT:
		{
			_service->OnPowerEvent(dwEventType, (PPOWERBROADCAST_SETTING)lpEventData);
			return NO_ERROR;
		}

		// the MSDN suggests returning NO_ERROR even if this event is not handled:
		case SERVICE_CONTROL_INTERROGATE:
			return NO_ERROR;
	}

	return ERROR_CALL_NOT_IMPLEMENTED;
}
